"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import logging
import re
from collections import ChainMap, deque
from collections.abc import Sequence
from copy import copy, deepcopy
from datetime import datetime, timedelta
from time import time
from typing import TYPE_CHECKING, Any

from bs4 import BeautifulSoup
from multidict import MultiDict
from typing_extensions import Self, TypeAlias
from yarl import URL as URL_

from . import utils
from .abc import BaseUser, Commentable, SteamID, UserDict
from .channel import DMChannel
from .clan import Clan
from .comment import Comment
from .enums import *
from .errors import *
from .game import GameToDict
from .group import Group
from .guard import *
from .invite import ClanInvite, UserInvite
from .iterators import AsyncIterator
from .message import *
from .message import ClanMessage
from .models import URL, EventParser, Registerable, register
from .protobufs import (
    EMsg,
    Msg,
    MsgProto,
    chat,
    client_server,
    client_server_2,
    comments,
    do_nothing_case,
    econ,
    friend_messages,
    friends,
    game_servers,
    login,
    player,
    struct_messages,
)
from .trade import DescriptionDict, TradeOffer, TradeOfferDict
from .user import User

if TYPE_CHECKING:
    from .abc import Message
    from .client import Client
    from .gateway import SteamWebSocket

log = logging.getLogger(__name__)


class TradeQueue:
    def __init__(self):
        self.queue: deque[TradeOffer] = deque()
        self._waiting_for: dict[int, asyncio.Future[TradeOffer]] = {}

    async def wait_for(self, id: int) -> TradeOffer:
        for trade in reversed(self.queue):  # check if it's already here
            if trade.id == id:
                self.queue.remove(trade)
                return trade

        self._waiting_for[id] = future = asyncio.get_running_loop().create_future()
        trade = await future
        self.queue.remove(trade)
        return trade

    def __len__(self) -> int:
        return len(self.queue)

    def __iadd__(self, other: list[TradeOffer]) -> Self:
        for trade in other:
            try:
                future = self._waiting_for[trade.id]
            except KeyError:
                pass
            else:
                future.set_result(trade)
                del self._waiting_for[trade.id]

        self.queue += other
        return self


class ConnectionState(Registerable):
    parsers: dict[EMsg, EventParser[Any]]

    def __init__(self, client: Client, **kwargs: Any):
        self.client = client
        self.dispatch = client.dispatch
        self.http = client.http

        self.handled_friends = asyncio.Event()
        self.handled_emoticons = asyncio.Event()
        self.handled_groups = asyncio.Event()
        self.handled_group_members = asyncio.Event()
        self.max_messages: int = kwargs.pop("max_messages", 1000)

        game = kwargs.get("game")
        games = kwargs.get("games")
        games = [game.to_dict() for game in games] if games is not None else []
        if game is not None:
            games.append(game.to_dict())
        self._games: list[GameToDict] = games
        self._state: PersonaState = kwargs.get("state", PersonaState.Online)
        self._ui_mode: UIMode = kwargs.get("ui_mode", UIMode.Desktop)
        self._flags: PersonaStateFlag = kwargs.get("flags", PersonaStateFlag.NONE)
        self._force_kick: bool = kwargs.get("force_kick", False)

        self.clear()

    def clear(self) -> None:
        self._users: dict[int, User] = {}
        self._trades: dict[int, TradeOffer] = {}
        self._groups: dict[int, Group] = {}
        self._clans: dict[int, Clan] = {}
        self._confirmations: dict[int, Confirmation] = {}
        self._confirmations_to_ignore: list[int] = []
        self._messages: deque[Message] = deque(maxlen=self.max_messages or 0)
        self.invites: dict[int, UserInvite | ClanInvite] = {}
        self.emoticons: list[ClientEmoticon] = []
        self.trade_queue = TradeQueue()
        self.previous_notification = None

        self._trades_to_watch: set[int] = set()
        self._trades_received_cache: Sequence[dict[str, Any]] = ()
        self._trades_sent_cache: Sequence[dict[str, Any]] = ()

        self.handled_friends.clear()
        self.handled_emoticons.clear()
        self.handled_groups.clear()

    async def __ainit__(self) -> None:
        if self.http.api_key is not None:
            self._device_id = generate_device_id(self.client.user)

            await self.poll_trades()

    @property
    def ws(self) -> SteamWebSocket:
        return self.client.ws

    @property
    def steam_time(self) -> datetime:
        if self.ws is None:  # can't be more precise
            return datetime.utcnow()
        return datetime.utcnow() + self.ws.server_offset

    @property
    def users(self) -> list[User]:
        return list(self._users.values())

    @property
    def trades(self) -> list[TradeOffer]:
        return list(self._trades.values())

    @property
    def groups(self) -> list[Group]:
        return list(self._groups.values())

    @property
    def clans(self) -> list[Clan]:
        return list(self._clans.values())

    @property
    def confirmations(self) -> list[Confirmation]:
        return list(self._confirmations.values())

    def get_user(self, id64: int) -> User | None:
        return self._users.get(id64)

    async def fetch_user(self, user_id64: int) -> User | None:
        data = await self.http.get_user(user_id64)
        return self._store_user(data) if data else None
        # return (await self.fetch_users([user_id64]))[0]

    async def fetch_users(self, user_id64s: list[int]) -> list[User | None]:
        resp = await self.http.get_users(user_id64s)
        """
        msg: MsgProto[CMsgClientRequestFriendData] = MsgProto(
            EMsg.ClientRequestFriendData,
            persona_state_requested=863,  # 1 | 2 | 4 | 8 | 16 | 64 | 256 | 512  corresponds to EClientPersonaStateFlag
            friends=user_id64s,
        )
        await self.ws.send_as_proto(msg)
        """

        return [self._store_user(data) for data in resp]

    async def _maybe_user(self, id64: int) -> User | SteamID:
        return self.get_user(id64) or await self.fetch_user(id64) or SteamID(id64)

    def _store_user(self, data: UserDict) -> User:
        try:
            user = self._users[int(data["steamid"])]
        except KeyError:
            user = User(state=self, data=data)
            self._users[user.id64] = user
        else:
            user._update(data)
        return user

    def get_confirmation(self, id: int) -> Confirmation | None:
        return self._confirmations.get(id)

    async def fetch_confirmation(self, id: int) -> Confirmation | None:
        await self._fetch_confirmations()
        return self.get_confirmation(id)

    def get_group(self, id: int) -> Group | None:
        return self._groups.get(id)

    def get_clan(self, id: int) -> Clan | None:
        return self._clans.get(id)

    async def fetch_clan(self, id64: int) -> Clan | None:
        msg: MsgProto[chat.GetClanChatRoomInfoResponse] = await self.ws.send_um_and_wait(
            "ClanChatRooms.GetClanChatRoomInfo", steamid=id64
        )
        if msg.result == Result.Busy:
            raise WSNotFound(msg)
        if msg.result != Result.OK:
            raise WSException(msg)

        return await Clan._from_proto(self, msg.body)

    def get_trade(self, id: int) -> TradeOffer | None:
        return self._trades.get(id)

    async def fetch_trade(self, id: int) -> TradeOffer | None:
        resp = await self.http.get_trade(id)
        if resp.get("response"):
            trade = [resp["response"]["offer"]]
            descriptions = resp["response"].get("descriptions", ())
            trades = await self._process_trades(trade, descriptions)
            return trades[0]

    async def _store_trade(self, data: TradeOfferDict) -> TradeOffer:
        try:
            trade = self._trades[int(data["tradeofferid"])]
        except KeyError:
            log.info(f'Received trade #{data["tradeofferid"]}')
            trade = TradeOffer._from_api(
                state=self, data=data, partner=await self._maybe_user(utils.make_id64(data["accountid_other"]))
            )
            self._trades[trade.id] = trade
            if trade.state in (TradeOfferState.Active, TradeOfferState.ConfirmationNeed) and (
                trade.items_to_send or trade.items_to_receive  # trade could be glitched
            ):
                self.dispatch("trade_send" if trade.is_our_offer() else "trade_receive", trade)
                self._trades_to_watch.add(trade.id)
        else:
            before_state = trade.state
            trade._update(data)
            if trade.state != before_state:
                log.info(f"Trade #{trade.id} has updated its trade state to {trade.state}")
                event_name = trade.state.event_name
                if event_name and (trade.items_to_send or trade.items_to_receive):
                    self.dispatch(f"trade_{event_name}", trade)
                    self._trades_to_watch.discard(trade.id)
        return trade

    async def _process_trades(
        self, trades: list[TradeOfferDict], descriptions: list[DescriptionDict]
    ) -> list[TradeOffer]:
        ret = []
        for trade in trades:
            for item in descriptions:
                for asset in trade.get("items_to_receive", ()):
                    if item["classid"] == asset["classid"] and item["instanceid"] == asset["instanceid"]:
                        asset.update(item)
                for asset in trade.get("items_to_give", ()):
                    if item["classid"] == asset["classid"] and item["instanceid"] == asset["instanceid"]:
                        asset.update(item)
            ret.append(await self._store_trade(trade))
        return ret

    @utils.call_once
    async def poll_trades(self) -> None:
        await self.fill_trades()

        while self._trades_to_watch:  # watch trades for changes
            await asyncio.sleep(5)
            await self.fill_trades()

    async def fill_trades(self) -> None:
        try:
            trades = await self.http.get_trade_offers()
            descriptions = trades.get("descriptions", ())
            trades_received = trades.get("trade_offers_received", ())
            trades_sent = trades.get("trade_offers_sent", ())
            updated_received_trades = [trade for trade in trades_received if trade not in self._trades_received_cache]
            updated_sent_trades = [trade for trade in trades_sent if trade not in self._trades_sent_cache]
            received_trades = await self._process_trades(updated_received_trades, descriptions)
            sent_trades = await self._process_trades(updated_sent_trades, descriptions)
            self._trades_received_cache = trades_received
            self._trades_sent_cache = trades_sent

            self.trade_queue += received_trades
            self.trade_queue += sent_trades
        except Exception as exc:
            await asyncio.sleep(30)
            log.info("Error while polling trades", exc_info=exc)

    async def wait_for_trade(self, id: int) -> TradeOffer:
        self.loop.create_task(self.poll_trades())  # start re-polling trades
        return await self.trade_queue.wait_for(id=id)

    # confirmations

    def _create_confirmation_params(self, tag: str) -> dict[str, Any]:
        timestamp = int(self.steam_time.timestamp())
        return {
            "p": self._device_id,
            "a": self.client.user.id64,
            "k": self._generate_confirmation(tag, timestamp),
            "t": timestamp,
            "m": "android",
            "tag": tag,
        }

    async def _fetch_confirmations(self) -> dict[int, Confirmation]:
        params = self._create_confirmation_params("conf")
        headers = {"X-Requested-With": "com.valvesoftware.android.steam.community"}
        resp = await self.http.get(URL.COMMUNITY / "mobileconf/conf", params=params, headers=headers)

        if "incorrect Steam Guard codes." in resp:
            raise InvalidCredentials("identity_secret is incorrect")

        soup = BeautifulSoup(resp, "html.parser")
        if soup.select("#mobileconf_empty"):
            return self._confirmations
        for confirmation in soup.select("#mobileconf_list .mobileconf_list_entry"):
            data_conf_id = confirmation["data-confid"]
            key = confirmation["data-key"]
            trade_id = int(confirmation.get("data-creator", 0))
            confirmation_id = confirmation["id"].split("conf")[1]
            if trade_id in self._confirmations_to_ignore:
                continue
            self._confirmations[trade_id] = Confirmation(self, confirmation_id, data_conf_id, key, trade_id)

        return self._confirmations

    def _generate_confirmation(self, tag: str, timestamp: int) -> str:
        return generate_confirmation_code(self.client.identity_secret, tag, timestamp)  # type: ignore

    async def fetch_and_confirm_confirmation(self, trade_id: int) -> bool:
        if self.client.identity_secret:
            confirmation = self.get_confirmation(trade_id) or await self.fetch_confirmation(trade_id)
            if confirmation is not None:
                await confirmation.confirm()
                return True

        return False

    # ws stuff

    @property
    def _combined(self) -> ChainMap[int, Group | Clan]:
        return ChainMap({clan.chat_id: clan for clan in self._clans.values()}, self._groups)  # type: ignore

    async def send_user_message(self, user_id64: int, content: str) -> UserMessage:
        contains_bbcode = utils.contains_bbcode(content)
        msg: MsgProto[friend_messages.SendMessageResponse] = await self.ws.send_um_and_wait(
            "FriendMessages.SendMessage",
            steamid=user_id64,
            message=content.replace("[", "\\[") if contains_bbcode else content,
            chat_entry_type=ChatEntryType.Text,
            contains_bbcode=contains_bbcode,
        )

        if msg.result == Result.LimitExceeded:
            raise WSForbidden(msg)
        if msg.result != Result.OK:
            raise WSException(msg)

        proto = friend_messages.IncomingMessageNotification(
            chat_entry_type=ChatEntryType.Text,
            message=content,
            rtime32_server_timestamp=int(time()),
            message_no_bbcode=msg.body.message_without_bb_code,
        )
        channel = DMChannel(state=self, participant=self.get_user(user_id64))  # type: ignore
        message = UserMessage(proto=proto, channel=channel)
        message.author = self.client.user
        self._messages.append(message)
        self.dispatch("message", message)

        return message

    async def send_user_typing(self, user_id64: int) -> None:
        await self.ws.send_um(
            "FriendMessages.SendMessage",
            steamid=user_id64,
            chat_entry_type=ChatEntryType.Typing,
        )
        self.dispatch("typing", self.client.user, datetime.utcnow())

    async def send_group_message(self, chat_id: int, group_id: int, content: str) -> ClanMessage | GroupMessage:
        msg: MsgProto[chat.SendChatMessageResponse] = await self.ws.send_um_and_wait(
            "ChatRoom.SendChatMessage", chat_id=chat_id, chat_group_id=group_id, message=content
        )

        if msg.result == Result.LimitExceeded:
            raise WSForbidden(msg)
        if msg.result == Result.InvalidParameter:
            raise WSNotFound(msg)
        if msg.result != Result.OK:
            raise WSException(msg)

        proto = chat.IncomingChatMessageNotification(
            chat_id=chat_id,
            chat_group_id=group_id,
            steamid_sender=0,
            message=content,
            message_no_bbcode=msg.body.message_without_bb_code,
            timestamp=int(time()),
        )
        group = self._combined[group_id]
        channel = group._channels[chat_id]
        channel._update(proto)

        message = (ClanMessage if isinstance(group, Clan) else GroupMessage)(
            proto=proto, channel=channel, author=self.client.user
        )
        self._messages.append(message)
        self.dispatch("message", message)

        return message

    async def join_chat(self, chat_id: int, invite_code: str | None = None) -> None:
        msg = await self.ws.send_um_and_wait(
            "ChatRoom.JoinChatRoomGroup", chat_group_id=chat_id, invite_code=invite_code or ""
        )

        if msg.result == Result.InvalidParameter:
            raise WSNotFound(msg)
        elif msg.result != Result.OK:
            raise WSException(msg)

    async def leave_chat(self, chat_id: int) -> None:
        msg = await self.ws.send_um_and_wait("ChatRoom.LeaveChatRoomGroup", chat_group_id=chat_id)

        if msg.result == Result.InvalidParameter:
            raise WSNotFound(msg)
        elif msg.result != Result.OK:
            raise WSException(msg)

    async def invite_user_to_group(self, user_id64: int, group_id: int) -> None:
        msg = await self.ws.send_um_and_wait(
            "ChatRoom.InviteFriendToChatRoomGroup", chat_group_id=group_id, steamid=user_id64
        )

        if msg.result == Result.InvalidParameter:
            raise WSNotFound(msg)
        elif msg.result != Result.OK:
            raise WSException(msg)

    async def edit_role(self, group_id: int, role_id: int, *, name: str) -> None:
        msg = await self.ws.send_um_and_wait("ChatRoom.RenameRole", chat_group_id=group_id, role_id=role_id, name=name)
        if msg.result == Result.InvalidParameter:
            raise WSNotFound(msg)
        elif msg.result != Result.OK:
            raise WSException(msg)

    async def fetch_user_history(
        self, user_id64: int, start: int, last: int
    ) -> friend_messages.GetRecentMessagesResponse:
        msg: MsgProto[friend_messages.GetRecentMessagesResponse] = await self.ws.send_um_and_wait(
            "FriendMessages.GetRecentMessages",
            steamid1=self.client.user.id64,
            steamid2=user_id64,
            rtime32_start_time=start,
            time_last=last,
            count=100,
        )

        if msg.result != Result.OK:
            raise WSException(msg)

        return msg.body

    async def fetch_group_history(
        self, group_id: int, chat_id: int, start: int, last: int
    ) -> chat.GetMessageHistoryResponse:
        msg: MsgProto[chat.GetMessageHistoryResponse] = await self.ws.send_um_and_wait(
            "ChatRoom.GetMessageHistory",
            chat_group_id=group_id,
            chat_id=chat_id,
            last_time=last,
            start_time=start,
            max_count=100,
        )

        if msg.result != Result.OK:
            raise WSException(msg)

        return msg.body

    async def fetch_servers(self, query: str, limit: int) -> list[game_servers.GetServerListResponseServer]:
        msg: MsgProto[game_servers.GetServerListResponse] = await self.ws.send_um_and_wait(
            "GameServers.GetServerList",
            filter=query,
            limit=limit,
        )

        if msg.result != Result.OK:
            raise WSException(msg)

        return msg.body.servers

    async def fetch_server_ip_from_steam_id(self, *ids: int) -> list[game_servers.IPsWithSteamIDsResponseServer]:
        msg: MsgProto[game_servers.IPsWithSteamIDsResponse] = await self.ws.send_um_and_wait(
            "GameServers.GetServerIPsBySteamID",
            server_steamids=list(ids),
        )

        if msg.result != Result.OK:
            raise WSException(msg)

        return msg.body.servers

    async def fetch_game_player_count(self, game_id: int) -> int:
        msg: MsgProto[client_server_2.CMsgDpGetNumberOfCurrentPlayersResponse] = await self.ws.send_proto_and_wait(
            MsgProto(EMsg.ClientGetNumberOfCurrentPlayersDP, appid=game_id)
        )
        if msg.result != Result.OK:
            raise WSException(msg)
        return msg.body.player_count

    # TODO use this in V.1
    async def fetch_user_games(self, user_id64: int) -> list[player.GetOwnedGamesResponseGame]:
        msg: MsgProto[player.GetOwnedGamesResponse] = await self.ws.send_um_and_wait(
            "Player.GetOwnedGames",
            steamid=user_id64,
            include_appinfo=True,
            include_played_free_games=True,
        )

        if msg.result != Result.OK:
            raise WSException(msg)

        return msg.body.games

    async def fetch_user_profile_info(self, user_id64: int) -> friends.CMsgClientFriendProfileInfoResponse:
        msg: MsgProto[friends.CMsgClientFriendProfileInfoResponse] = await self.ws.send_proto_and_wait(
            MsgProto(EMsg.ClientFriendProfileInfo, steamid_friend=user_id64)
        )
        if msg.result != Result.OK:
            raise WSException(msg)

        return msg.body

    async def fetch_user_equipped_profile_items(self, user_id64: int) -> player.GetProfileItemsEquippedResponse:
        msg: MsgProto[player.GetProfileItemsEquippedResponse] = await self.ws.send_um_and_wait(
            "Player.GetProfileItemsEquipped",
            steamid=user_id64,
        )
        if msg.result != Result.OK:
            raise WSException(msg)

        return msg.body

    async def fetch_profile_items(self) -> player.GetProfileItemsOwnedResponse:
        msg: MsgProto[player.GetProfileItemsOwnedResponse] = await self.ws.send_um_and_wait(
            "Player.GetProfileItemsOwned",
        )
        if msg.result != Result.OK:
            raise WSException(msg)

        return msg.body

    async def fetch_user_favourite_badge(self, user_id64) -> player.GetFavoriteBadgeResponse:
        msg: MsgProto[player.GetFavoriteBadgeResponse] = await self.ws.send_um_and_wait(
            "Player.GetFavoriteBadge",
            steamid=user_id64,
        )
        if msg.result != Result.OK:
            raise WSException(msg)

        return msg.body

    async def fetch_trade_url(self, generate_new: bool) -> str:
        msg: MsgProto[econ.GetTradeOfferAccessTokenResponse] = await self.ws.send_um_and_wait(
            "Econ.GetTradeOfferAccessToken", generate_new_token=generate_new
        )
        if msg.result != Result.OK:
            raise WSException(msg)
        return (
            f"https://steamcommunity.com/tradeoffer/new/?partner={self.client.user.id}"
            f"&token={msg.body.trade_offer_access_token}"
        )

    # parsers

    @register(EMsg.ServiceMethod)
    async def parse_service_method(self, msg: MsgProto[Any]) -> None:
        name = msg.header.body.job_name_target
        if name == "ChatRoomClient.NotifyIncomingChatMessage":
            await self.handle_group_message(msg)
        elif name == "ChatRoomClient.NotifyChatRoomHeaderStateChange":
            await self.handle_group_update(msg)
        elif name == "ChatRoomClient.NotifyChatGroupUserStateChanged":
            await self.handle_group_user_action(msg)
        elif name == "ChatRoomClient.NotifyChatRoomGroupRoomsChange":
            await self.handle_group_channel_update(msg)
        elif name == "FriendMessagesClient.IncomingMessage":
            await self.handle_user_message(msg)
        else:
            log.debug("Got an event %r that we don't handle %r", msg.header.body.job_name_target, msg.body)

    async def handle_user_message(self, msg: MsgProto[friend_messages.IncomingMessageNotification]) -> None:
        partner = await self._maybe_user(msg.body.steamid_friend)  # FIXME shouldn't ever be out of cache
        author = self.client.user if msg.body.local_echo else partner  # local_echo is always us

        if msg.body.chat_entry_type == ChatEntryType.Text:
            channel = DMChannel(state=self, participant=partner)
            message = UserMessage(proto=msg.body, channel=channel)
            message.author = author
            self._messages.append(message)
            self.dispatch("message", message)

        if msg.body.chat_entry_type == ChatEntryType.Typing:
            when = datetime.utcfromtimestamp(msg.body.rtime32_server_timestamp)
            self.dispatch("typing", author, when)

    async def handle_group_message(self, msg: MsgProto[chat.IncomingChatMessageNotification]) -> None:
        try:
            destination = self._combined[msg.body.chat_group_id]
        except KeyError:
            return log.debug(f"Got a message for a chat we aren't in {msg.body.chat_group_id}")

        channel = destination._channels[msg.body.chat_id]
        channel._update(msg.body)
        author = await self._maybe_user(msg.body.steamid_sender)
        message = (ClanMessage if isinstance(destination, Clan) else GroupMessage)(
            proto=msg.body, channel=channel, author=author
        )
        self._messages.append(message)
        self.dispatch("message", message)

    async def handle_group_update(self, msg: MsgProto[chat.ChatRoomHeaderStateNotification]) -> None:
        try:
            group = self._combined[msg.body.header_state.chat_group_id]
        except KeyError:
            return log.debug(f"Updating a group that isn't cached {msg.body.header_state.chat_group_id}")

        before = copy(group)
        before._channels = {c_id: copy(c) for c_id, c in before._channels.items()}
        before.roles = deepcopy(before.roles)
        if isinstance(before, Group):
            before.members = before.members.copy()
        group._update(msg.body.header_state)
        self.dispatch(f"{'group' if isinstance(destination, Group) else 'clan'}_update", before, group)

    async def handle_group_user_action(self, msg: MsgProto[chat.NotifyChatGroupUserStateChangedNotification]) -> None:
        if msg.body.user_action == "Joined":  # join group
            if msg.body.group_summary.clanid:
                clan = await Clan._from_proto(self, msg.body.group_summary)
                self._clans[clan.id] = clan
                self.dispatch("clan_join", clan)
            else:
                group = await Group._from_proto(self, msg.body.group_summary)
                self._groups[group.id] = group
                self.dispatch("group_join", group)

        elif msg.body.user_action == "Parted":  # leave group
            left = self._combined.pop(msg.body.chat_group_id, None)
            if left is None:
                return

            if isinstance(left, Clan):
                self.dispatch("clan_leave", left)
            else:
                self.dispatch("group_leave", left)

    async def handle_group_channel_update(self, msg: MsgProto[chat.ChatRoomGroupRoomsChangeNotification]):
        try:
            group = self._combined[msg.body.chat_group_id]
        except KeyError:
            return log.debug(f"Got an update for a clan we aren't in {msg.body.chat_group_id}")

        before = copy(group)
        before._channels = {c_id: copy(c) for c_id, c in before._channels.items()}
        before.roles = deepcopy(before.roles)
        if isinstance(before, Group):
            before.members = before.members.copy()
        group._update(msg.body)
        self.dispatch(f"{'clan' if isinstance(group, Clan) else 'group'}_update", before, group)

    @register(EMsg.ServiceMethodResponse)
    async def parse_service_method_response(self, msg: MsgProto[Any]) -> None:
        if msg.header.body.job_name_target == "ChatRoom.GetMyChatRoomGroups":
            msg: MsgProto[chat.GetMyChatRoomGroupsResponse]
            for group in msg.body.chat_room_groups:
                if group.group_summary.clanid:  # received a clan
                    clan = await Clan._from_proto(self, group)
                    self._clans[clan.id] = clan
                else:  # else it's a group
                    group = await Group._from_proto(self, group.group_summary)
                    self._groups[group.id] = group
            self.handled_groups.set()  # signal to process_group_members that we are ready
            await self.handled_group_members.wait()  # ensure the members are ready
            await self.handled_friends.wait()  # ensure friend cache is ready
            # await self.handled_emoticons.wait()  # ensure emoticon cache is ready
            await self.client._handle_ready()

    @register(EMsg.ClientCMList)
    async def parse_cm_list_update(self, msg: MsgProto[client_server.CMsgClientCmList]) -> None:
        log.debug("Updating CM list")
        cms = msg.body.cm_websocket_addresses
        cm_list = self.ws.cm_list
        cm_list.extend(cms)
        await cm_list.ping()  # ping all the cms, we have time.

    @register(EMsg.ClientPersonaState)
    async def parse_persona_state_update(self, msg: MsgProto[friends.CMsgClientPersonaState]) -> None:
        for friend in msg.body.friends:
            data: UserDict = friend.to_dict(do_nothing_case)
            if not data:
                continue
            user_id64 = friend.friendid
            after = self.get_user(user_id64)
            if after is None:  # they're private
                continue

            before = copy(after)

            try:
                data = self.patch_user_from_ws(data, friend)
            except (KeyError, TypeError):
                invitee = await self._maybe_user(user_id64)
                invite = UserInvite(self, invitee, FriendRelationship.RequestRecipient)
                self.dispatch("user_invite", invite)

            after._update(data)
            old = [getattr(before, attr, None) for attr in BaseUser.__slots__]
            new = [getattr(after, attr, None) for attr in BaseUser.__slots__]
            if old != new:
                self.dispatch("user_update", before, after)

    def patch_user_from_ws(self, data: dict[str, Any], friend: friends.CMsgClientPersonaStateFriend) -> dict:
        data["personaname"] = friend.player_name
        hash = (
            friend.avatar_hash.hex()
            if data["avatar_hash"] != "\x00" * 20
            else "fef49e7fa7e1997310d705b2a6158ff8dc1cdfeb"
        )
        data[
            "avatarfull"
        ] = f"https://steamcdn-a.akamaihd.net/steamcommunity/public/images/avatars/{hash[:2]}/{hash}_full.jpg"

        if friend.last_logoff:
            data["lastlogoff"] = friend.last_logoff
        data["gameextrainfo"] = friend.game_name or None
        data["personastate"] = friend.persona_state
        data["personastateflags"] = friend.persona_state_flags
        return data

    @register(EMsg.ClientFriendsList)
    async def process_friends(self, msg: MsgProto[friends.CMsgClientFriendsList]) -> None:
        elements = None
        client_user_friends: list[int] = []
        is_load = not msg.body.bincremental
        for friend in msg.body.friends:
            relationship = FriendRelationship.try_value(friend.efriendrelationship)
            steam_id = SteamID(friend.ulfriendid)

            if relationship == FriendRelationship.Friend:
                try:
                    invite = self.invites.pop(steam_id.id64)
                except KeyError:
                    if is_load:
                        client_user_friends.append(steam_id.id64)
                else:
                    self.dispatch(f"{'user'if steam_id.type == Type.Individual else 'clan'}_invite_accept", invite)
            elif relationship in (
                FriendRelationship.RequestInitiator,
                FriendRelationship.RequestRecipient,
            ):
                if steam_id.type == Type.Individual:
                    invitee = await self._maybe_user(steam_id.id64)
                    invite = UserInvite(state=self, invitee=invitee, relationship=relationship)
                    self.invites[invitee.id64] = invite
                    self.dispatch("user_invite", invite)
                if steam_id.type == Type.Clan:
                    if elements is None:
                        resp = await self.http.get(URL.COMMUNITY / "my/groups/pending", params={"ajax": "1"})
                        soup = BeautifulSoup(resp, "html.parser")
                        elements = soup.find_all("a", class_="linkStandard")
                    invitee_id64 = 0
                    for idx, element in enumerate(elements):
                        if str(steam_id.id64) in str(element):
                            invitee_id64 = utils.make_id64(elements[idx + 1]["data-miniprofile"])
                            break
                    invitee = await self._maybe_user(invitee_id64)
                    try:
                        clan = await self.fetch_clan(steam_id.id64) or steam_id
                    except WSException:
                        clan = steam_id
                    invite = ClanInvite(state=self, invitee=invitee, clan=clan, relationship=relationship)
                    self.invites[clan.id64] = invite
                    self.dispatch("clan_invite", invite)

            elif relationship == FriendRelationship.NONE and steam_id.type == Type.Individual:
                try:
                    invite = self.invites.pop(steam_id.id64)
                except KeyError:
                    friend = self.get_user(steam_id.id64)
                    try:
                        self.client.user.friends.remove(friend)
                    except ValueError:
                        pass
                    self.dispatch("user_remove", friend)
                else:
                    self.dispatch(f"{'user'if steam_id.type == Type.Individual else 'clan'}_invite_decline", invite)
        if is_load:
            self.client.user.friends = await self.fetch_users(client_user_friends)
            self.handled_friends.set()

    @register(EMsg.ClientFriendsGroupsList)
    async def process_group_members(self, msg: MsgProto[friends.CMsgClientFriendsGroupsList]):
        await self.handled_groups.wait()
        members = await self.fetch_users([m.ul_steam_id for m in msg.body.memberships])
        for membership in msg.body.memberships:
            for member in members:
                if member and member.id64 == membership.ul_steam_id:
                    try:
                        self._groups[membership.n_group_id].members.append(member)
                    except KeyError:
                        log.debug(f"Somehow got an unknown group {membership.n_group_id} and member {member.id64}")

        self.handled_group_members.set()

    async def fetch_group_roles(self, group_id: int) -> list[chat.Role]:
        msg: MsgProto[chat.GetRolesResponse] = await self.ws.send_um_and_wait(
            "ChatRoom.GetRoles", chat_group_id=group_id
        )
        if msg.result == Result.AccessDenied:
            raise WSForbidden(msg)
        if msg.result != Result.OK:
            raise WSException(msg)
        return msg.body.roles

    async def fetch_comment(self, owner: Commentable, id: int) -> comments.GetCommentThreadRatingsResponse.Comment:
        msg: MsgProto[comments.GetCommentThreadRatingsResponse] = await self.ws.send_um_and_wait(
            "Community.GetCommentThread", **owner._commentable_kwargs, id=id
        )
        if msg.result != Result.OK:
            raise WSException(msg)
        return msg.body.comments[0]

    async def fetch_comments(
        self, owner: Commentable, limit: int | None, after: datetime, oldest_first: bool
    ) -> list[comments.GetCommentThreadResponse.Comment]:
        msg: MsgProto[comments.GetCommentThreadResponse] = await self.ws.send_um_and_wait(
            "Community.GetCommentThread",
            **owner._commentable_kwargs,
            count=2147483647 if limit is None else limit,
            oldest_first=oldest_first,
            time_oldest=int(after.timestamp()),
        )  # int(i32::MAX / 2) not entirely sure why this is the max and not i32 which is the max for the field
        if msg.result != Result.OK:
            raise WSException(msg)
        return msg.body.comments

    async def post_comment(self, owner: Commentable, content: str, subscribe: bool) -> Comment[Commentable]:
        msg: MsgProto[comments.PostCommentToThreadResponse] = await self.ws.send_um_and_wait(
            "Community.PostCommentToThread",
            **owner._commentable_kwargs,
            content=content,
            surpress_notifications=not subscribe,
        )
        if msg.result != Result.OK:
            raise WSException(msg)

        comment = Comment(
            self,
            id=msg.body.id,
            content=content,
            created_at=self.steam_time,
            author=self.client.user,
            owner=owner,
        )
        self.dispatch("comment", comment)
        return comment

    async def delete_comment(self, owner: Commentable, comment_id: int) -> None:
        msg: MsgProto[comments.DeleteCommentFromThreadResponse] = await self.ws.send_um_and_wait(
            "Community.DeleteCommentFromThread",
            **owner._commentable_kwargs,
            id=comment_id,
        )
        if msg.result != Result.OK:
            raise WSException(msg)

    async def report_comment(self, owner: Commentable, comment_id: int) -> None:
        msg: MsgProto[comments.PostCommentToThreadResponse] = await self.ws.send_um_and_wait(
            "Community.PostCommentToThread",  # some odd api here
            **owner._commentable_kwargs,
            is_report=True,
            parent_id=comment_id,
        )
        if msg.result != Result.OK:
            raise WSException(msg)

    @register(EMsg.ClientCommentNotifications)
    async def handle_comments(self, msg: MsgProto[client_server_2.CMsgClientCommentNotifications]) -> None:
        resp = await self.http.get(URL.COMMUNITY / "my/commentnotifications")
        soup = BeautifulSoup(resp, "html.parser")

        cached: dict[URL_, list[Commentable | int]] = {}
        for attr in soup.find_all("div", class_="commentnotification_click_overlay", limit=msg.body.count_new_comments):
            url = URL_(attr.contents[1]["href"])
            url_with_no_query = url.with_query(None)
            if url_with_no_query in cached:
                cached[url_with_no_query][1] += 1  # type: ignore
                commentable, index = cached[url_with_no_query]  # type: ignore
                commentable: Commentable
                index: int
            else:
                steam_id = await SteamID.from_url(url, self.http._session)
                if steam_id is None:
                    continue
                if steam_id.type == Type.Individual:
                    commentable = self.get_user(steam_id.id64) or await self.fetch_user(steam_id.id64)
                else:
                    clan = self.get_clan(steam_id.id64) or await self.fetch_clan(steam_id.id64)
                    if clan is None:
                        continue
                    # now that we have the clan that the comment was posted in, if the url has a hash in the path we can
                    # extract the type of the comment section the comment is in.
                    if "#" not in url.path:
                        commentable = clan
                    else:
                        path, _, end = url.path.partition("#")[-1].partition("/")
                        if path == "events":
                            commentable = await clan.fetch_event(re.findall(r"([0-9]*)", end)[0])
                        elif path == "announcements":
                            commentable = await clan.fetch_announcement(re.findall(r"detail/([0-9]*)", end)[0])
                        else:
                            log.debug(f"Got a comment for a type we cannot handle {path}")
                            continue
                if commentable is None:
                    continue
                index = 0
                cached[url_with_no_query] = [commentable, index]

            try:
                timestamp = int(url.query["tscn"])
            except KeyError:
                log.debug("Got a comment without a timestamp")
                continue

            after = datetime.utcfromtimestamp(timestamp) - timedelta(minutes=1)

            comments = [comment async for comment in commentable.comments(limit=index + 1, after=after)]

            try:
                self.dispatch("comment", comments[index])
            except IndexError:
                pass

    @register(EMsg.ClientUserNotifications)
    async def parse_notification(self, msg: MsgProto[client_server_2.CMsgClientUserNotifications]) -> None:
        if msg.body != self.previous_notification and any(
            b.user_notification_type == 1 for b in msg.body.notifications
        ):
            # 1 is a trade offer
            await self.poll_trades()
        self.previous_notification = msg.body

    @register(EMsg.ClientItemAnnouncements)
    async def parse_new_items(self, msg: MsgProto[client_server_2.CMsgClientItemAnnouncements]) -> None:
        if msg.body.count_new_items:
            await self.poll_trades()

    @register(EMsg.ClientAccountInfo)
    def parse_account_info(self, msg: MsgProto[login.CMsgClientAccountInfo]) -> None:
        if self.client.user is None:
            return
        if msg.body.persona_name != self.client.user.name:
            before = copy(self.client.user)
            self.client.user.name = msg.body.persona_name or self.client.user.name
            self.dispatch("user_update", before, self.client.user)

    async def fetch_friends_who_own(self, game_id: int) -> list[int]:
        msg: Msg[struct_messages.ClientGetFriendsWhoPlayGameResponse] = await self.ws.send_proto_and_wait(
            Msg(EMsg.ClientGetFriendsWhoPlayGame, app_id=game_id)
        )
        if msg.result != Result.OK:
            raise WSException(msg)
        return msg.body.friends

    async def rate_clan_announcement(self, clan_id: int, announcement_id: int, upvoted: bool) -> None:
        msg = await self.ws.send_um_and_wait(
            "Community.RateClanAnnouncement",
            announcementid=announcement_id,
            clan_accountid=clan_id,
            vote_up=upvoted,
        )
        if msg.result != Result.OK:
            raise WSException(msg)

    async def respond_to_clan_invite(self, clan_id64: int, accept: bool) -> None:
        msg = await self.ws.send_um_and_wait("Clan.RespondToClanInvite", steamid=clan_id64, accept=accept)
        if msg.result != Result.OK:
            raise WSException(msg)

    @register(EMsg.ClientClanState)
    async def update_clan(self, msg: MsgProto[client_server.CMsgClientClanState]) -> None:
        await self.handled_groups.wait()
        clan = self.get_clan(msg.body.steamid_clan) or await self.fetch_clan(msg.body.steamid_clan)
        if clan is None:
            return

        for event in msg.body.events:
            if event.just_posted:
                event = await clan.fetch_event(event.gid)
                self.dispatch("event_create", event)
        for announcement in msg.body.announcements:
            if announcement.just_posted:
                announcement = await clan.fetch_announcement(announcement.gid)
                self.dispatch("announcement_create", announcement)

        user_counts = msg.body.user_counts
        name_info = msg.body.name_info
        if user_counts or name_info:
            before = copy(clan)
        if user_counts:
            clan.member_count = user_counts.members
            clan.in_game_count = user_counts.in_game
            clan.online_count = user_counts.online
            clan.active_member_count = user_counts.chatting
        if name_info:
            clan.name = name_info.clan_name
            hexed = name_info.sha_avatar.hex()
            hash = hexed if hexed != "\x00" * 20 else "fef49e7fa7e1997310d705b2a6158ff8dc1cdfeb"
            clan.avatar_url = (
                f"https://steamcdn-a.akamaihd.net/steamcommunity/public/images/avatars/{hash[:2]}/{hash}_full.jpg"
            )

        if user_counts or name_info:
            self.dispatch("clan_update", before, clan)


class TradesList(AsyncIterator[TradeOffer]):
    @utils.call_once
    async def fill(self) -> None:
        state = self._state
        try:
            trades = await state.http.get_trade_offers()
            descriptions = trades.get("descriptions", ())
            trades_received = trades.get("trade_offers_received", ())
            trades_sent = trades.get("trade_offers_sent", ())
            new_received_trades = [trade for trade in trades_received if trade not in state._trades_received_cache]
            new_sent_trades = [trade for trade in trades_sent if trade not in state._trades_sent_cache]
            received_trades = await state._process_trades(new_received_trades, descriptions)
            sent_trades = await state._process_trades(new_sent_trades, descriptions)
            state._trades_received_cache = trades_received
            state._trades_sent_cache = trades_sent

            self.queue += received_trades
            self.queue += sent_trades
        except Exception as exc:
            await asyncio.sleep(30)
            log.info("Error while polling trades", exc_info=exc)

    async def wait_for(self, id: int) -> TradeOffer:
        if not self.queue:
            await self.fill()  # if this is a no-op it doesn't matter
        trade = utils.get(self.queue, id=id)
        if trade is not None:
            self.queue.remove(trade)
            return trade

        await self.fill()  # refresh the queue
        return await self.wait_for(id=id)



"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import abc
from collections.abc import Sequence
from datetime import datetime
from typing import TYPE_CHECKING, Any, overload

from typing_extensions import Literal

from . import utils
from .abc import Commentable, SteamID
from .enums import ClanEvent
from .game import Game, StatefulGame

if TYPE_CHECKING:
    from .clan import Clan
    from .game_server import GameServer
    from .state import ConnectionState
    from .user import User

__all__ = (
    "Event",
    "Announcement",
)


class BaseEvent(Commentable, utils.AsyncInit, metaclass=abc.ABCMeta):
    __slots__ = (
        "clan",
        "id",
        "author",
        "name",
        "content",
        "game",
        "starts_at",
        "becomes_visible",
        "stops_being_visible",
        "end",
        "type",
        "last_edited_at",
        "last_edited_by",
        "hidden",
        "published",
        "upvotes",
        "downvotes",
        "comment_count",
        "server_address",
        "server_password",
        "_state",
    )

    # data here is of type steammessages_base.CClanEventData.to_dict()
    # TODO keep checking if there is a way to get CClanEventData directly from ws
    def __init__(self, state: ConnectionState, clan: Clan, data: dict[str, Any]):
        self._state = state
        self.clan = clan
        self.id: int = int(data["gid"])
        self.author: User | SteamID | None = int(data["creator_steamid"])  # type: ignore
        self.name: str = data["event_name"]
        self.content: str = data["event_notes"]
        self.game = StatefulGame(state, id=data["appid"]) if data["appid"] else None
        self.starts_at = datetime.utcfromtimestamp(data["rtime32_start_time"])
        self.becomes_visible = (
            datetime.utcfromtimestamp(data["rtime32_visibility_start"])
            if data.get("rtime32_visibility_start")
            else None
        )
        self.stops_being_visible = (
            datetime.utcfromtimestamp(data["rtime32_visibility_end"]) if data.get("rtime32_visibility_end") else None
        )
        self.end = datetime.utcfromtimestamp(data["rtime32_end_time"]) if data["rtime32_end_time"] else None
        self.type = ClanEvent.try_value(data["event_type"])
        self.last_edited_at = datetime.utcfromtimestamp(data["rtime32_last_modified"])
        self.last_edited_by: User | SteamID | None = (  # type: ignore
            int(data["last_update_steamid"]) if (data["last_update_steamid"] or "0") != "0" else None
        )
        self.hidden = bool(data.get("hidden", 0))
        self.published = bool(data.get("published", 1))
        self.upvotes: int = data.get("votes_up", 0)
        self.downvotes: int = data.get("votes_down", 0)
        self.comment_count: int = data.get("comment_count", 0)
        self.server_address: str | None = data["server_address"]
        self.server_password: str | None = data["server_password"]

        self._feature = int(data["gidfeature"])
        self._feature2 = int(data.get("gidfeature2", 0) or 0)

    async def __ainit__(self) -> None:
        self.author = await self._state._maybe_user(self.author)  # type: ignore
        if self.last_edited_by:
            self.last_edited_by = await self._state._maybe_user(self.last_edited_by)  # type: ignore

    def __repr__(self) -> str:
        attrs = ("id", "name", "type", "author", "clan")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"


class Event(BaseEvent):
    """Represents an event in a clan.

    Attributes
    ----------
    id
        The event's id.
    author
        The event's author.
    name
        The event's name.
    content
        The event's content.
    game
        The game the event is going to play in.
    clan
        The event's clan.
    start
        The event's start time.
    becomes_visible
        The time at which the event becomes visible.
    stops_being_visible
        The time at which the event stops being visible.
    end
        The time at which the event ends.
    type
        The event's type.
    last_edited_at
        The time the event was last edited at.
    last_edited_by
        The the user who made the event last edited.
    hidden
        Whether or not the event is currently hidden.
    published
        Whether or not the event is currently published.
    upvotes
        The number of up votes on the event.
    downvotes
        The number of down votes on the event.
    comment_count
        The number of comments on the event.
    server_address
        The event's server address.
    server_password
        The event's server password.
    """

    server_address: str
    server_password: str

    @property
    def _commentable_kwargs(self) -> dict[str, Any]:
        return {
            "thread_type": 14,
            "id64": self.clan.id64,
            "gidfeature": self._feature,
        }

    async def server(self) -> GameServer | None:
        """The server that the game will be ran on, ``None`` if not found.

        Note
        ----
        This is shorthand for

        .. code-block:: python3

            await client.fetch_server(ip=event.server_address)
        """

        if self.server_address == "0":
            return
        return await self._state.client.fetch_server(ip=self.server_address)

    @overload
    async def edit(
        self,
        name: str,
        content: str,
        *,
        type: Literal[
            ClanEvent.Other,
            ClanEvent.Chat,
            ClanEvent.Party,
            ClanEvent.Meeting,
            ClanEvent.SpecialCause,
            ClanEvent.MusicAndArts,
            ClanEvent.Sports,
            ClanEvent.Trip,
        ] = ClanEvent.Other,
        start: datetime | None = None,
    ) -> None:
        ...

    @overload
    async def edit(
        self,
        name: str,
        content: str,
        *,
        game: Game,
        type: Literal[ClanEvent.Game] = ...,
        start: datetime | None = ...,
        server_address: str | None = ...,
        server_password: str | None = ...,
    ) -> Event:
        ...

    async def edit(
        self,
        name: str,
        content: str,
        *,
        type: Literal[
            ClanEvent.Other,
            ClanEvent.Chat,
            ClanEvent.Game,
            # ClanEvent.Broadcast,  # TODO need to wait until implementing stream support for this
            ClanEvent.Party,
            ClanEvent.Meeting,
            ClanEvent.SpecialCause,
            ClanEvent.MusicAndArts,
            ClanEvent.Sports,
            ClanEvent.Trip,
        ] = ClanEvent.Other,
        game: Game | None = None,
        start: datetime | None = None,
        server_address: str | None = None,
        server_password: str | None = None,
    ) -> None:
        """Edit the event's details.

        Note
        ----
        If parameters are omitted they use what they are currently set to.

        Parameters
        ----------
        name
            The event's name.
        content
            The event's content.
        type
            The event's type.
        game
            The event's game.
        start
            The event's start time.
        server_address
            The event's server's address.
        server_password
            The event's server's password.
        """
        type = type or self.type
        new_game = game or self.game
        game_id = str(new_game) if new_game is not None else None
        await self._state.http.edit_clan_event(
            self.clan.id64,
            name or self.name,
            content or self.content,
            f"{type.name}Event",
            game_id or "",
            server_address or self.server_address if self.server_address else "",
            server_password or self.server_password if self.server_password else "",
            start or self.start,
            event_id=self.id,
        )
        self.name = name or self.name
        self.content = content or self.content
        self.type = type or self.type
        self.server_address = server_address or self.server_address
        self.server_password = server_password or self.server_password
        self.game = StatefulGame(self._state, id=game_id) if game_id is not None else None
        self.last_edited_at = datetime.utcnow()
        self.last_edited_by = self._state.client.user

    async def delete(self) -> None:
        await self._state.http.delete_clan_event(self.clan.id64, self.id)


class Announcement(BaseEvent):
    """Represents an announcement in a clan.

    Attributes
    ----------
    id
        The announcement's id.
    author
        The announcement's author.
    name
        The announcement's name.
    content
        The announcement's content.
    game
        The game the announcement is for.
    clan
        The announcement's clan.
    start
        The announcement's start time.
    becomes_visible
        The time at which the announcement becomes visible.
    stops_being_visible
        The time at which the announcement stops being visible.
    end
        The time at which the announcement ends.
    type
        The announcement's type.
    last_edited_at
        The time the announcement was last edited at.
    last_edited_by
        The the user who made the announcement last edited.
    hidden
        Whether or not the announcement is currently hidden.
    published
        Whether or not the announcement is currently published.
    upvotes
        The number of up votes on the announcement.
    downvotes
        The number of down votes on the announcement.
    comment_count
        The number of comments on the announcement.
    topic_id
        The id of the forum post comments are sent to.
    created_at
        The time at which the announcement was created at.
    updated_at
        The time at which the announcement was last updated_at.
    approved_at
        The time at which the announcement was approved by a moderator.
    tags
        The announcement's tags.
    """

    __slots__ = (
        "topic_id",
        "created_at",
        "updated_at",
        "approved_at",
        "tags",
    )
    approved_by: User
    server_ip: None
    server_password: None

    def __init__(self, state: ConnectionState, clan: Clan, data: dict[str, Any]):
        super().__init__(state, clan, data)
        body: dict[str, Any] = data["announcement_body"]
        self.id: int = int(body["gid"])
        self.topic_id = int(data["forum_topic_id"]) if data["forum_topic_id"] else None
        self.created_at = datetime.utcfromtimestamp(body["posttime"])
        self.updated_at = datetime.utcfromtimestamp(body["updatetime"])  # this is different to self.edited_at?
        self.approved_at = datetime.utcfromtimestamp(data["rtime_mod_reviewed"]) if data["rtime_mod_reviewed"] else None
        self.content: str = body["body"]
        self.tags: Sequence[str] = body["tags"]

    @property
    def _commentable_kwargs(self) -> dict[str, Any]:
        if self.clan.is_game_clan:
            raise NotImplementedError("Fetching a game announcement's comments is not currently supported")
        return {
            "thread_type": 13,
            "id64": self.clan.id64,
            "gidfeature": self._feature,
        }

    async def edit(self, name: str | None = None, content: str | None = None) -> None:
        """Edit the announcement's details.

        Note
        ----
        If parameters are omitted they use what they are currently set to.

        Parameters
        ----------
        name
            The announcement's name.
        content
            The announcement's content.
        """
        name = name or self.name
        content = content or self.content
        await self._state.http.edit_clan_announcement(self.clan.id64, self.id, name, content)
        self.name = name
        self.content = content
        self.last_edited_at = datetime.utcnow()
        self.last_edited_by = self._state.client.user

    async def hide(self) -> None:
        """Hide this announcement."""
        await self._state.http.hide_clan_announcement(self.clan.id, self.id, True)

    async def unhide(self) -> None:
        """Un-hide this announcement."""
        await self._state.http.hide_clan_announcement(self.clan.id, self.id, False)

    async def upvote(self) -> None:
        """Upvote this announcement."""
        return await self._state.rate_clan_announcement(self.clan.id, self.id, True)

    async def downvote(self) -> None:
        """Downvote this announcement."""
        return await self._state.rate_clan_announcement(self.clan.id, self.id, False)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_clientserver_2.proto
# plugin: python-betterproto
# Last updated 09/09/2021

from dataclasses import dataclass
from typing import List

import betterproto


@dataclass(eq=False, repr=False)
class CMsgClientUpdateUserGameInfo(betterproto.Message):
    steamid_idgs: int = betterproto.fixed64_field(1)
    gameid: int = betterproto.fixed64_field(2)
    game_ip: int = betterproto.uint32_field(3)
    game_port: int = betterproto.uint32_field(4)
    token: bytes = betterproto.bytes_field(5)


@dataclass(eq=False, repr=False)
class CMsgClientRichPresenceUpload(betterproto.Message):
    rich_presence_kv: bytes = betterproto.bytes_field(1)
    steamid_broadcast: List[int] = betterproto.fixed64_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientRichPresenceRequest(betterproto.Message):
    steamid_request: List[int] = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientRichPresenceInfo(betterproto.Message):
    rich_presence: List["CMsgClientRichPresenceInfoRichPresence"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientRichPresenceInfoRichPresence(betterproto.Message):
    steamid_user: int = betterproto.fixed64_field(1)
    rich_presence_kv: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientCheckFileSignature(betterproto.Message):
    app_id: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientCheckFileSignatureResponse(betterproto.Message):
    app_id: int = betterproto.uint32_field(1)
    pid: int = betterproto.uint32_field(2)
    eresult: int = betterproto.uint32_field(3)
    filename: str = betterproto.string_field(4)
    esignatureresult: int = betterproto.uint32_field(5)
    sha_file: bytes = betterproto.bytes_field(6)
    signatureheader: bytes = betterproto.bytes_field(7)
    filesize: int = betterproto.uint32_field(8)
    getlasterror: int = betterproto.uint32_field(9)
    evalvesignaturecheckdetail: int = betterproto.uint32_field(10)


@dataclass(eq=False, repr=False)
class CMsgClientReadMachineAuth(betterproto.Message):
    filename: str = betterproto.string_field(1)
    offset: int = betterproto.uint32_field(2)
    cubtoread: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientReadMachineAuthResponse(betterproto.Message):
    filename: str = betterproto.string_field(1)
    eresult: int = betterproto.uint32_field(2)
    filesize: int = betterproto.uint32_field(3)
    sha_file: bytes = betterproto.bytes_field(4)
    getlasterror: int = betterproto.uint32_field(5)
    offset: int = betterproto.uint32_field(6)
    cubread: int = betterproto.uint32_field(7)
    bytes_read: bytes = betterproto.bytes_field(8)
    filename_sentry: str = betterproto.string_field(9)


@dataclass(eq=False, repr=False)
class CMsgClientUpdateMachineAuth(betterproto.Message):
    filename: str = betterproto.string_field(1)
    offset: int = betterproto.uint32_field(2)
    cubtowrite: int = betterproto.uint32_field(3)
    bytes_: bytes = betterproto.bytes_field(4)
    otp_type: int = betterproto.uint32_field(5)
    otp_identifier: str = betterproto.string_field(6)
    otp_sharedsecret: bytes = betterproto.bytes_field(7)
    otp_timedrift: int = betterproto.uint32_field(8)


@dataclass(eq=False, repr=False)
class CMsgClientUpdateMachineAuthResponse(betterproto.Message):
    filename: str = betterproto.string_field(1)
    eresult: int = betterproto.uint32_field(2)
    filesize: int = betterproto.uint32_field(3)
    sha_file: bytes = betterproto.bytes_field(4)
    getlasterror: int = betterproto.uint32_field(5)
    offset: int = betterproto.uint32_field(6)
    cubwrote: int = betterproto.uint32_field(7)
    otp_type: int = betterproto.int32_field(8)
    otp_value: int = betterproto.uint32_field(9)
    otp_identifier: str = betterproto.string_field(10)


@dataclass(eq=False, repr=False)
class CMsgClientRequestMachineAuth(betterproto.Message):
    filename: str = betterproto.string_field(1)
    eresult_sentryfile: int = betterproto.uint32_field(2)
    filesize: int = betterproto.uint32_field(3)
    sha_sentryfile: bytes = betterproto.bytes_field(4)
    lock_account_action: int = betterproto.int32_field(6)
    otp_type: int = betterproto.uint32_field(7)
    otp_identifier: str = betterproto.string_field(8)
    otp_sharedsecret: bytes = betterproto.bytes_field(9)
    otp_value: int = betterproto.uint32_field(10)
    machine_name: str = betterproto.string_field(11)
    machine_name_userchosen: str = betterproto.string_field(12)


@dataclass(eq=False, repr=False)
class CMsgClientRequestMachineAuthResponse(betterproto.Message):
    eresult: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientRegisterKey(betterproto.Message):
    key: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientPurchaseResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    purchaseresult_details: int = betterproto.int32_field(2)
    purchase_receipt_info: bytes = betterproto.bytes_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientActivateOemLicense(betterproto.Message):
    bios_manufacturer: str = betterproto.string_field(1)
    bios_serialnumber: str = betterproto.string_field(2)
    license_file: bytes = betterproto.bytes_field(3)
    mainboard_manufacturer: str = betterproto.string_field(4)
    mainboard_product: str = betterproto.string_field(5)
    mainboard_serialnumber: str = betterproto.string_field(6)


@dataclass(eq=False, repr=False)
class CMsgClientRegisterOemMachine(betterproto.Message):
    oem_register_file: bytes = betterproto.bytes_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientRegisterOemMachineResponse(betterproto.Message):
    eresult: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientPurchaseWithMachineId(betterproto.Message):
    package_id: int = betterproto.uint32_field(1)
    machine_info: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class CMsgTradingInitiateTradeRequest(betterproto.Message):
    trade_request_id: int = betterproto.uint32_field(1)
    other_steamid: int = betterproto.uint64_field(2)
    other_name: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class CMsgTradingInitiateTradeResponse(betterproto.Message):
    response: int = betterproto.uint32_field(1)
    trade_request_id: int = betterproto.uint32_field(2)
    other_steamid: int = betterproto.uint64_field(3)
    steamguard_required_days: int = betterproto.uint32_field(4)
    new_device_cooldown_days: int = betterproto.uint32_field(5)
    default_password_reset_probation_days: int = betterproto.uint32_field(6)
    password_reset_probation_days: int = betterproto.uint32_field(7)
    default_email_change_probation_days: int = betterproto.uint32_field(8)
    email_change_probation_days: int = betterproto.uint32_field(9)


@dataclass(eq=False, repr=False)
class CMsgTradingCancelTradeRequest(betterproto.Message):
    other_steamid: int = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class CMsgTradingStartSession(betterproto.Message):
    other_steamid: int = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientGetCdnAuthToken(betterproto.Message):
    depot_id: int = betterproto.uint32_field(1)
    host_name: str = betterproto.string_field(2)
    app_id: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientGetDepotDecryptionKey(betterproto.Message):
    depot_id: int = betterproto.uint32_field(1)
    app_id: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientGetDepotDecryptionKeyResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    depot_id: int = betterproto.uint32_field(2)
    depot_encryption_key: bytes = betterproto.bytes_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientCheckAppBetaPassword(betterproto.Message):
    app_id: int = betterproto.uint32_field(1)
    betapassword: str = betterproto.string_field(2)
    language: int = betterproto.int32_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientCheckAppBetaPasswordResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    betapasswords: List["CMsgClientCheckAppBetaPasswordResponseBetaPassword"] = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class CMsgClientCheckAppBetaPasswordResponseBetaPassword(betterproto.Message):
    betaname: str = betterproto.string_field(1)
    betapassword: str = betterproto.string_field(2)
    betadescription: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientUpdateAppJobReport(betterproto.Message):
    app_id: int = betterproto.uint32_field(1)
    depot_ids: List[int] = betterproto.uint32_field(2)
    app_state: int = betterproto.uint32_field(3)
    job_app_error: int = betterproto.uint32_field(4)
    error_details: str = betterproto.string_field(5)
    job_duration: int = betterproto.uint32_field(6)
    files_validation_failed: int = betterproto.uint32_field(7)
    job_bytes_downloaded: int = betterproto.uint64_field(8)
    job_bytes_staged: int = betterproto.uint64_field(9)
    bytes_comitted: int = betterproto.uint64_field(10)
    start_app_state: int = betterproto.uint32_field(11)
    stats_machine_id: int = betterproto.fixed64_field(12)
    branch_name: str = betterproto.string_field(13)
    total_bytes_downloaded: int = betterproto.uint64_field(14)
    total_bytes_staged: int = betterproto.uint64_field(15)
    total_bytes_restored: int = betterproto.uint64_field(16)
    is_borrowed: bool = betterproto.bool_field(17)
    is_free_weekend: bool = betterproto.bool_field(18)
    total_bytes_legacy: int = betterproto.uint64_field(19)
    total_bytes_patched: int = betterproto.uint64_field(20)
    total_bytes_saved: int = betterproto.uint64_field(21)
    cell_id: int = betterproto.uint32_field(22)


@dataclass(eq=False, repr=False)
class CMsgClientDpContentStatsReport(betterproto.Message):
    stats_machine_id: int = betterproto.fixed64_field(1)
    country_code: str = betterproto.string_field(2)
    os_type: int = betterproto.int32_field(3)
    language: int = betterproto.int32_field(4)
    num_install_folders: int = betterproto.uint32_field(5)
    num_installed_games: int = betterproto.uint32_field(6)
    size_installed_games: int = betterproto.uint64_field(7)


@dataclass(eq=False, repr=False)
class CMsgClientGetCdnAuthTokenResponse(betterproto.Message):
    eresult: int = betterproto.uint32_field(1)
    token: str = betterproto.string_field(2)
    expiration_time: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class CMsgDownloadRateStatistics(betterproto.Message):
    cell_id: int = betterproto.uint32_field(1)
    stats: List["CMsgDownloadRateStatisticsStatsInfo"] = betterproto.message_field(2)
    throttling_kbps: int = betterproto.uint32_field(3)
    steam_realm: int = betterproto.uint32_field(4)


@dataclass(eq=False, repr=False)
class CMsgDownloadRateStatisticsStatsInfo(betterproto.Message):
    source_type: int = betterproto.uint32_field(1)
    source_id: int = betterproto.uint32_field(2)
    seconds: int = betterproto.uint32_field(3)
    bytes_: int = betterproto.uint64_field(4)
    host_name: str = betterproto.string_field(5)
    microseconds: int = betterproto.uint64_field(6)
    used_ipv6: bool = betterproto.bool_field(7)
    proxied: bool = betterproto.bool_field(8)


@dataclass(eq=False, repr=False)
class CMsgClientRequestAccountData(betterproto.Message):
    account_or_email: str = betterproto.string_field(1)
    action: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientRequestAccountDataResponse(betterproto.Message):
    action: int = betterproto.uint32_field(1)
    eresult: int = betterproto.uint32_field(2)
    account_name: str = betterproto.string_field(3)
    ct_matches: int = betterproto.uint32_field(4)
    account_name_suggestion1: str = betterproto.string_field(5)
    account_name_suggestion2: str = betterproto.string_field(6)
    account_name_suggestion3: str = betterproto.string_field(7)


@dataclass(eq=False, repr=False)
class CMsgClientUgsGetGlobalStats(betterproto.Message):
    gameid: int = betterproto.uint64_field(1)
    history_days_requested: int = betterproto.uint32_field(2)
    time_last_requested: int = betterproto.fixed32_field(3)
    first_day_cached: int = betterproto.uint32_field(4)
    days_cached: int = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class CMsgClientUgsGetGlobalStatsResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    timestamp: int = betterproto.fixed32_field(2)
    day_current: int = betterproto.int32_field(3)
    days: List["CMsgClientUgsGetGlobalStatsResponseDay"] = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class CMsgClientUgsGetGlobalStatsResponseDay(betterproto.Message):
    day_id: int = betterproto.uint32_field(1)
    stats: List["CMsgClientUgsGetGlobalStatsResponseDayStat"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientUgsGetGlobalStatsResponseDayStat(betterproto.Message):
    stat_id: int = betterproto.int32_field(1)
    data: int = betterproto.int64_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientRedeemGuestPass(betterproto.Message):
    guest_pass_id: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientRedeemGuestPassResponse(betterproto.Message):
    eresult: int = betterproto.uint32_field(1)
    package_id: int = betterproto.uint32_field(2)
    must_own_appid: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientGetClanActivityCounts(betterproto.Message):
    steamid_clans: List[int] = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientGetClanActivityCountsResponse(betterproto.Message):
    eresult: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientOgsReportString(betterproto.Message):
    accumulated: bool = betterproto.bool_field(1)
    sessionid: int = betterproto.uint64_field(2)
    severity: int = betterproto.int32_field(3)
    formatter: str = betterproto.string_field(4)
    varargs: bytes = betterproto.bytes_field(5)


@dataclass(eq=False, repr=False)
class CMsgClientOgsReportBug(betterproto.Message):
    sessionid: int = betterproto.uint64_field(1)
    bugtext: str = betterproto.string_field(2)
    screenshot: bytes = betterproto.bytes_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientSentLogs(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CMsgGcClient(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    msgtype: int = betterproto.uint32_field(2)
    payload: bytes = betterproto.bytes_field(3)
    steamid: int = betterproto.fixed64_field(4)
    gcname: str = betterproto.string_field(5)
    ip: int = betterproto.uint32_field(6)


@dataclass(eq=False, repr=False)
class CMsgClientRequestFreeLicense(betterproto.Message):
    appids: List[int] = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientRequestFreeLicenseResponse(betterproto.Message):
    eresult: int = betterproto.uint32_field(1)
    granted_packageids: List[int] = betterproto.uint32_field(2)
    granted_appids: List[int] = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class CMsgDrmDownloadRequestWithCrashData(betterproto.Message):
    download_flags: int = betterproto.uint32_field(1)
    download_types_known: int = betterproto.uint32_field(2)
    guid_drm: bytes = betterproto.bytes_field(3)
    guid_split: bytes = betterproto.bytes_field(4)
    guid_merge: bytes = betterproto.bytes_field(5)
    module_name: str = betterproto.string_field(6)
    module_path: str = betterproto.string_field(7)
    crash_data: bytes = betterproto.bytes_field(8)


@dataclass(eq=False, repr=False)
class CMsgDrmDownloadResponse(betterproto.Message):
    eresult: int = betterproto.uint32_field(1)
    app_id: int = betterproto.uint32_field(2)
    blob_download_type: int = betterproto.uint32_field(3)
    merge_guid: bytes = betterproto.bytes_field(4)
    download_file_dfs_ip: int = betterproto.uint32_field(5)
    download_file_dfs_port: int = betterproto.uint32_field(6)
    download_file_url: str = betterproto.string_field(7)
    module_path: str = betterproto.string_field(8)


@dataclass(eq=False, repr=False)
class CMsgDrmFinalResult(betterproto.Message):
    eresult: int = betterproto.uint32_field(1)
    app_id: int = betterproto.uint32_field(2)
    blob_download_type: int = betterproto.uint32_field(3)
    error_detail: int = betterproto.uint32_field(4)
    merge_guid: bytes = betterproto.bytes_field(5)
    download_file_dfs_ip: int = betterproto.uint32_field(6)
    download_file_dfs_port: int = betterproto.uint32_field(7)
    download_file_url: str = betterproto.string_field(8)


@dataclass(eq=False, repr=False)
class CMsgClientDpCheckSpecialSurvey(betterproto.Message):
    survey_id: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientDpCheckSpecialSurveyResponse(betterproto.Message):
    eresult: int = betterproto.uint32_field(1)
    state: int = betterproto.uint32_field(2)
    name: str = betterproto.string_field(3)
    custom_url: str = betterproto.string_field(4)
    include_software: bool = betterproto.bool_field(5)
    token: bytes = betterproto.bytes_field(6)


@dataclass(eq=False, repr=False)
class CMsgClientDpSendSpecialSurveyResponse(betterproto.Message):
    survey_id: int = betterproto.uint32_field(1)
    data: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientDpSendSpecialSurveyResponseReply(betterproto.Message):
    eresult: int = betterproto.uint32_field(1)
    token: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientRequestForgottenPasswordEmail(betterproto.Message):
    account_name: str = betterproto.string_field(1)
    password_tried: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientRequestForgottenPasswordEmailResponse(betterproto.Message):
    eresult: int = betterproto.uint32_field(1)
    use_secret_question: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientItemAnnouncements(betterproto.Message):
    count_new_items: int = betterproto.uint32_field(1)
    unseen_items: List["CMsgClientItemAnnouncementsUnseenItem"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientItemAnnouncementsUnseenItem(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    context_id: int = betterproto.uint64_field(2)
    asset_id: int = betterproto.uint64_field(3)
    amount: int = betterproto.uint64_field(4)
    rtime32_gained: int = betterproto.fixed32_field(5)
    source_appid: int = betterproto.uint32_field(6)


@dataclass(eq=False, repr=False)
class CMsgClientRequestItemAnnouncements(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CMsgClientUserNotifications(betterproto.Message):
    notifications: List["CMsgClientUserNotificationsNotification"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientUserNotificationsNotification(betterproto.Message):
    user_notification_type: int = betterproto.uint32_field(1)
    count: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientCommentNotifications(betterproto.Message):
    count_new_comments: int = betterproto.uint32_field(1)
    count_new_comments_owner: int = betterproto.uint32_field(2)
    count_new_comments_subscriptions: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientRequestCommentNotifications(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CMsgClientOfflineMessageNotification(betterproto.Message):
    offline_messages: int = betterproto.uint32_field(1)
    friends_with_offline_messages: List[int] = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientRequestOfflineMessageCount(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CMsgClientChatGetFriendMessageHistory(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientChatGetFriendMessageHistoryResponse(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    success: int = betterproto.uint32_field(2)
    messages: List["CMsgClientChatGetFriendMessageHistoryResponseFriendMessage"] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientChatGetFriendMessageHistoryResponseFriendMessage(betterproto.Message):
    accountid: int = betterproto.uint32_field(1)
    timestamp: int = betterproto.uint32_field(2)
    message: str = betterproto.string_field(3)
    unread: bool = betterproto.bool_field(4)


@dataclass(eq=False, repr=False)
class CMsgClientChatGetFriendMessageHistoryForOfflineMessages(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CMsgClientFsGetFriendsSteamLevels(betterproto.Message):
    accountids: List[int] = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientFsGetFriendsSteamLevelsResponse(betterproto.Message):
    friends: List["CMsgClientFsGetFriendsSteamLevelsResponseFriend"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientFsGetFriendsSteamLevelsResponseFriend(betterproto.Message):
    accountid: int = betterproto.uint32_field(1)
    level: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientEmailAddrInfo(betterproto.Message):
    email_address: str = betterproto.string_field(1)
    email_is_validated: bool = betterproto.bool_field(2)
    email_validation_changed: bool = betterproto.bool_field(3)
    credential_change_requires_code: bool = betterproto.bool_field(4)
    password_or_secretqa_change_requires_code: bool = betterproto.bool_field(5)
    remind_user_about_email: bool = betterproto.bool_field(6)


@dataclass(eq=False, repr=False)
class CMsgCreItemVoteSummary(betterproto.Message):
    published_file_ids: List["CMsgCreItemVoteSummaryPublishedFileId"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class CMsgCreItemVoteSummaryPublishedFileId(betterproto.Message):
    published_file_id: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class CMsgCreItemVoteSummaryResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    item_vote_summaries: List["CMsgCreItemVoteSummaryResponseItemVoteSummary"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class CMsgCreItemVoteSummaryResponseItemVoteSummary(betterproto.Message):
    published_file_id: int = betterproto.fixed64_field(1)
    votes_for: int = betterproto.int32_field(2)
    votes_against: int = betterproto.int32_field(3)
    reports: int = betterproto.int32_field(4)
    score: float = betterproto.float_field(5)


@dataclass(eq=False, repr=False)
class CMsgCreUpdateUserPublishedItemVote(betterproto.Message):
    published_file_id: int = betterproto.fixed64_field(1)
    vote_up: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class CMsgCreUpdateUserPublishedItemVoteResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)


@dataclass(eq=False, repr=False)
class CMsgCreGetUserPublishedItemVoteDetails(betterproto.Message):
    published_file_ids: List["CMsgCreGetUserPublishedItemVoteDetailsPublishedFileId"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class CMsgCreGetUserPublishedItemVoteDetailsPublishedFileId(betterproto.Message):
    published_file_id: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class CMsgCreGetUserPublishedItemVoteDetailsResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    user_item_vote_details: List[
        "CMsgCreGetUserPublishedItemVoteDetailsResponseUserItemVoteDetail"
    ] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class CMsgCreGetUserPublishedItemVoteDetailsResponseUserItemVoteDetail(betterproto.Message):
    published_file_id: int = betterproto.fixed64_field(1)
    vote: int = betterproto.int32_field(2)


@dataclass(eq=False, repr=False)
class CMsgFsGetFollowerCount(betterproto.Message):
    steam_id: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class CMsgFsGetFollowerCountResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    count: int = betterproto.int32_field(2)


@dataclass(eq=False, repr=False)
class CMsgFsGetIsFollowing(betterproto.Message):
    steam_id: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class CMsgFsGetIsFollowingResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    is_following: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class CMsgFsEnumerateFollowingList(betterproto.Message):
    start_index: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CMsgFsEnumerateFollowingListResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    total_results: int = betterproto.int32_field(2)
    steam_ids: List[int] = betterproto.fixed64_field(3)


@dataclass(eq=False, repr=False)
class CMsgDpGetNumberOfCurrentPlayers(betterproto.Message):
    appid: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CMsgDpGetNumberOfCurrentPlayersResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    player_count: int = betterproto.int32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientFriendUserStatusPublished(betterproto.Message):
    friend_steamid: int = betterproto.fixed64_field(1)
    appid: int = betterproto.uint32_field(2)
    status_text: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientServiceMethodLegacy(betterproto.Message):
    method_name: str = betterproto.string_field(1)
    serialized_method: bytes = betterproto.bytes_field(2)
    is_notification: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientServiceMethodLegacyResponse(betterproto.Message):
    method_name: str = betterproto.string_field(1)
    serialized_method_response: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientUiMode(betterproto.Message):
    uimode: int = betterproto.uint32_field(1)
    chat_mode: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientVanityUrlChangedNotification(betterproto.Message):
    vanity_url: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientAuthorizeLocalDeviceRequest(betterproto.Message):
    device_description: str = betterproto.string_field(1)
    owner_account_id: int = betterproto.uint32_field(2)
    local_device_token: int = betterproto.uint64_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientAuthorizeLocalDevice(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    owner_account_id: int = betterproto.uint32_field(2)
    authed_device_token: int = betterproto.uint64_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientAuthorizeLocalDeviceNotification(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    owner_account_id: int = betterproto.uint32_field(2)
    local_device_token: int = betterproto.uint64_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientDeauthorizeDeviceRequest(betterproto.Message):
    deauthorization_account_id: int = betterproto.uint32_field(1)
    deauthorization_device_token: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientDeauthorizeDevice(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    deauthorization_account_id: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientUseLocalDeviceAuthorizations(betterproto.Message):
    authorization_account_id: List[int] = betterproto.uint32_field(1)
    device_tokens: List["CMsgClientUseLocalDeviceAuthorizationsDeviceToken"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientUseLocalDeviceAuthorizationsDeviceToken(betterproto.Message):
    owner_account_id: int = betterproto.uint32_field(1)
    token_id: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientGetAuthorizedDevices(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CMsgClientGetAuthorizedDevicesResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    authorized_device: List["CMsgClientGetAuthorizedDevicesResponseAuthorizedDevice"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientGetAuthorizedDevicesResponseAuthorizedDevice(betterproto.Message):
    auth_device_token: int = betterproto.uint64_field(1)
    device_name: str = betterproto.string_field(2)
    last_access_time: int = betterproto.uint32_field(3)
    borrower_id: int = betterproto.uint32_field(4)
    is_pending: bool = betterproto.bool_field(5)
    app_played: int = betterproto.uint32_field(6)


@dataclass(eq=False, repr=False)
class CMsgClientSharedLibraryLockStatus(betterproto.Message):
    locked_library: List["CMsgClientSharedLibraryLockStatusLockedLibrary"] = betterproto.message_field(1)
    own_library_locked_by: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientSharedLibraryLockStatusLockedLibrary(betterproto.Message):
    owner_id: int = betterproto.uint32_field(1)
    locked_by: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientSharedLibraryStopPlaying(betterproto.Message):
    seconds_left: int = betterproto.int32_field(1)
    stop_apps: List["CMsgClientSharedLibraryStopPlayingStopApp"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientSharedLibraryStopPlayingStopApp(betterproto.Message):
    app_id: int = betterproto.uint32_field(1)
    owner_id: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientServiceCall(betterproto.Message):
    sysid_routing: bytes = betterproto.bytes_field(1)
    call_handle: int = betterproto.uint32_field(2)
    module_crc: int = betterproto.uint32_field(3)
    module_hash: bytes = betterproto.bytes_field(4)
    function_id: int = betterproto.uint32_field(5)
    cub_output_max: int = betterproto.uint32_field(6)
    flags: int = betterproto.uint32_field(7)
    callparameter: bytes = betterproto.bytes_field(8)
    ping_only: bool = betterproto.bool_field(9)
    max_outstanding_calls: int = betterproto.uint32_field(10)


@dataclass(eq=False, repr=False)
class CMsgClientServiceModule(betterproto.Message):
    module_crc: int = betterproto.uint32_field(1)
    module_hash: bytes = betterproto.bytes_field(2)
    module_content: bytes = betterproto.bytes_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientServiceCallResponse(betterproto.Message):
    sysid_routing: bytes = betterproto.bytes_field(1)
    call_handle: int = betterproto.uint32_field(2)
    module_crc: int = betterproto.uint32_field(3)
    module_hash: bytes = betterproto.bytes_field(4)
    ecallresult: int = betterproto.uint32_field(5)
    result_content: bytes = betterproto.bytes_field(6)
    os_version_info: bytes = betterproto.bytes_field(7)
    system_info: bytes = betterproto.bytes_field(8)
    load_address: int = betterproto.fixed64_field(9)
    exception_record: bytes = betterproto.bytes_field(10)
    portable_os_version_info: bytes = betterproto.bytes_field(11)
    portable_system_info: bytes = betterproto.bytes_field(12)
    was_converted: bool = betterproto.bool_field(13)
    internal_result: int = betterproto.uint32_field(14)
    current_count: int = betterproto.uint32_field(15)
    last_call_handle: int = betterproto.uint32_field(16)
    last_call_module_crc: int = betterproto.uint32_field(17)
    last_call_sysid_routing: bytes = betterproto.bytes_field(18)
    last_ecallresult: int = betterproto.uint32_field(19)
    last_callissue_delta: int = betterproto.uint32_field(20)
    last_callcomplete_delta: int = betterproto.uint32_field(21)


@dataclass(eq=False, repr=False)
class CMsgAmUnlockStreaming(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CMsgAmUnlockStreamingResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    encryption_key: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class CMsgAmUnlockHevc(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CMsgAmUnlockHevcResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientPlayingSessionState(betterproto.Message):
    playing_blocked: bool = betterproto.bool_field(2)
    playing_app: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientKickPlayingSession(betterproto.Message):
    only_stop_game: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientVoiceCallPreAuthorize(betterproto.Message):
    caller_steamid: int = betterproto.fixed64_field(1)
    receiver_steamid: int = betterproto.fixed64_field(2)
    caller_id: int = betterproto.int32_field(3)
    hangup: bool = betterproto.bool_field(4)


@dataclass(eq=False, repr=False)
class CMsgClientVoiceCallPreAuthorizeResponse(betterproto.Message):
    caller_steamid: int = betterproto.fixed64_field(1)
    receiver_steamid: int = betterproto.fixed64_field(2)
    eresult: int = betterproto.int32_field(3)
    caller_id: int = betterproto.int32_field(4)


@dataclass(eq=False, repr=False)
class CMsgBadgeCraftedNotification(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    badge_level: int = betterproto.uint32_field(2)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_contentsystem.steamclient.proto
# plugin: python-betterproto
# Last updated 09/09/2021

from dataclasses import dataclass
from typing import List

import betterproto


@dataclass(eq=False, repr=False)
class GetServersForSteamPipeRequest(betterproto.Message):
    cell_id: int = betterproto.uint32_field(1)
    max_servers: int = betterproto.uint32_field(2)
    ip_override: str = betterproto.string_field(3)
    launcher_type: int = betterproto.int32_field(4)
    ipv6_public: str = betterproto.string_field(5)


@dataclass(eq=False, repr=False)
class ServerInfo(betterproto.Message):
    type: str = betterproto.string_field(1)
    source_id: int = betterproto.int32_field(2)
    cell_id: int = betterproto.int32_field(3)
    load: int = betterproto.int32_field(4)
    weighted_load: float = betterproto.float_field(5)
    num_entries_in_client_list: int = betterproto.int32_field(6)
    steam_china_only: bool = betterproto.bool_field(7)
    host: str = betterproto.string_field(8)
    vhost: str = betterproto.string_field(9)
    use_as_proxy: bool = betterproto.bool_field(10)
    proxy_request_path_template: str = betterproto.string_field(11)
    https_support: str = betterproto.string_field(12)
    allowed_app_ids: List[int] = betterproto.uint32_field(13)
    preferred_server: bool = betterproto.bool_field(14)


@dataclass(eq=False, repr=False)
class GetServersForSteamPipeResponse(betterproto.Message):
    servers: List["ServerInfo"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetDepotPatchInfoRequest(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    depotid: int = betterproto.uint32_field(2)
    source_manifestid: int = betterproto.uint64_field(3)
    target_manifestid: int = betterproto.uint64_field(4)


@dataclass(eq=False, repr=False)
class GetDepotPatchInfoResponse(betterproto.Message):
    is_available: bool = betterproto.bool_field(1)
    patch_size: int = betterproto.uint64_field(2)
    patched_chunks_size: int = betterproto.uint64_field(3)


@dataclass(eq=False, repr=False)
class GetClientUpdateHostsRequest(betterproto.Message):
    cached_signature: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetClientUpdateHostsResponse(betterproto.Message):
    hosts_kv: str = betterproto.string_field(1)
    valid_until_time: int = betterproto.uint64_field(2)
    ip_country: str = betterproto.string_field(3)
